#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 University of Illinois
from typing import List


class CellExecution:
    """
    A cell execution (object) corresponds to a cell execution (action, i.e. press play) in the notebook session.
    """

    def __init__(
        self,
        cell_num: int,
        cell: str,
        cell_runtime: float,
        start_time: float,
        src_vss: List,
        dst_vss: List,
    ):
        """
        Create an operation event from cell execution metrics.
        Args:
            cell_num (int): The nth cell execution of the current session.
            cell (str): Raw cell cell.
            cell_runtime (float): Cell runtime.
            start_time (time): Time of start of cell execution. Note that this is different from when the cell was
                queued.
            src_vss (List[VariableSnapshot]): Nodeset containing input VSs of the cell execution.
            dst_vss (List[VariableSnapshot]): Nodeset containing output VSs of the cell execution.
        """
        self.cell_num = cell_num
        self.cell = cell
        self.cell_runtime = cell_runtime
        self.start_time = start_time

        self.src_vss = src_vss
        self.dst_vss = dst_vss



import time
from typing import Dict

import numpy as np
from ipykernel.zmqshell import ZMQInteractiveShell

from elastic.algorithm.selector import Selector
from elastic.core.common.profile_variable_size import profile_variable_size
from elastic.core.graph.graph import DependencyGraph
from elastic.core.io.migrate import migrate
from elastic.core.mutation.object_hash import UnserializableObj


def checkpoint(
    graph: DependencyGraph,
    shell: ZMQInteractiveShell,
    fingerprint_dict: Dict,
    selector: Selector,
    udfs: set,
    filename: str,
    profile_dict,
    write_log_location=None,
    notebook_name=None,
    optimizer_name=None,
    elastic_notebook=None,
):
    """
    Checkpoints the notebook. The optimizer selects the VSs to migrate and recompute and the OEs to recompute, then
    writes the checkpoint as the specified filename.
    Args:
        graph (DependencyGraph): dependency graph representation of the notebook.
        shell (ZMQInteractiveShell): interactive Jupyter shell storing the state of the current session.
        selector (Selector): optimizer for computing the checkpointing configuration.
        udfs (set): set of user-declared functions.
        filename (str): location to write the file to.
        write_log_location (str): location to write component runtimes to. For experimentation only.
        notebook_name (str): notebook name. For experimentation only.
        optimizer_name (str): optimizer name. For experimentation only.
        elastic_notebook (ElasticNotebook, optional): ElasticNotebook instance to update migration lists.
    """
    profile_start = time.time()

    # Retrieve active VSs from the graph. Active VSs are correspond to the latest instances/versions of each variable.
    active_vss = set()
    print("---------------------------")
    print("all variables:")
    # print(graph.variable_snapshots)
    for vs_list in graph.variable_snapshots.values():
        if not vs_list[-1].deleted:
            print(f"name: {vs_list[-1].name}")
            active_vss.add(vs_list[-1])

    # Profile the size of each variable defined in the current session.
    for active_vs in active_vss:
        # 変数がfingerprint_dictに存在するかチェック
        if active_vs.name not in fingerprint_dict:
            print(f"Warning: Variable '{active_vs.name}' not found in fingerprint_dict")
            continue

        attr_str = getattr(shell.user_ns[active_vs.name], "__module__", None)
        # Object is unserializable
        if isinstance(fingerprint_dict[active_vs.name][2], UnserializableObj):
            active_vs.size = np.inf

        # Blacklisted object
        elif attr_str and ("dataprep.eda" in attr_str or "bokeh" in attr_str):
            active_vs.size = np.inf

        # Profile size of object.
        else:
            active_vs.size = profile_variable_size(shell.user_ns[active_vs.name])

    # Check for pairwise variable intersections. Variables sharing underlying data must be migrated or recomputed
    # together.
    overlapping_vss = []
    for active_vs1 in active_vss:
        for active_vs2 in active_vss:
            if active_vs1 != active_vs2:
                # 両方の変数がfingerprint_dictに存在するかチェック
                if active_vs1.name not in fingerprint_dict or active_vs2.name not in fingerprint_dict:
                    continue
                if fingerprint_dict[active_vs1.name][1].intersection(fingerprint_dict[active_vs2.name][1]):
                    overlapping_vss.append((active_vs1, active_vs2))

    profile_end = time.time()
    if write_log_location:
        with open(
            write_log_location
            + "/output_"
            + notebook_name
            + "_"
            + optimizer_name
            + ".txt",
            "a",
        ) as f:
            f.write("overlappings - " + repr(len(overlapping_vss)) + "\n")
            f.write(
                "Profile stage took - "
                + repr(profile_start - profile_end)
                + " seconds"
                + "\n"
            )
            f.write(
                "Idgraph stage took - "
                + repr(profile_dict["idgraph"])
                + " seconds"
                + "\n"
            )
            f.write(
                "Representation stage took - "
                + repr(profile_dict["representation"])
                + " seconds"
                + "\n"
            )

    optimize_start = time.time()
    # Initialize the optimizer.
    add_start = time.time()
    selector.dependency_graph = graph
    selector.active_vss = active_vss
    selector.overlapping_vss = overlapping_vss
    add_end = time.time()

    # Use the optimizer to compute the checkpointing configuration.
    opt_start = time.time()
    vss_to_migrate, ces_to_recompute = selector.select_vss(
        write_log_location, notebook_name, optimizer_name
    )
    opt_end = time.time()
    print("---------------------------")
    print("variables to migrate:")
    for vs in vss_to_migrate:
        print(f"name: {vs.name}, size: {vs.size}")

    difference_start = time.time()
    vss_to_recompute = active_vss - vss_to_migrate
    difference_end = time.time()

    print("---------------------------")
    print("variables to recompute:")
    for vs in vss_to_recompute:
        print(f"name: {vs.name}, size: {vs.size}")
    print([vs.name for vs in vss_to_recompute])

    print("---------------------------")
    print("cells to recompute:")
    for ce in ces_to_recompute:
        print(f"cell num: {ce.cell_num}, cell runtime: {ce.cell_runtime}")
    print(sorted([ce.cell_num + 1 for ce in ces_to_recompute]))

    optimize_end = time.time()
    if write_log_location:
        with open(
            write_log_location
            + "/output_"
            + notebook_name
            + "_"
            + optimizer_name
            + ".txt",
            "a",
        ) as f:
            f.write(
                "Optimize stage took - "
                + repr(optimize_end - optimize_start)
                + " seconds"
                + "\n"
            )
            f.write(
                "  Add stage took - " + repr(add_end - add_start) + " seconds" + "\n"
            )
            f.write(
                "  Opt stage took - " + repr(opt_end - opt_start) + " seconds" + "\n"
            )
            f.write(
                "  Diff stage took - "
                + repr(difference_end - difference_start)
                + " seconds"
                + "\n"
            )

    # 変数リストを更新
    if elastic_notebook is not None:
        elastic_notebook.update_migration_lists(vss_to_migrate, vss_to_recompute)

    # Store the notebook checkpoint to the specified location.
    migrate_start = time.time()
    migrate_success = True
    migrate(
        graph,
        shell,
        vss_to_migrate,
        vss_to_recompute,
        ces_to_recompute,
        udfs,
        selector.recomputation_ces,
        selector.overlapping_vss,
        filename,
    )
    migrate_end = time.time()

    if write_log_location:
        with open(
            write_log_location
            + "/output_"
            + notebook_name
            + "_"
            + optimizer_name
            + ".txt",
            "a",
        ) as f:
            f.write(
                "Migrate stage took - "
                + repr(migrate_end - migrate_start)
                + " seconds"
                + "\n"
            )
    return migrate_success



from __future__ import print_function

import sys
import time
import traceback
import types
from os.path import dirname

from IPython import get_ipython
from IPython.core.interactiveshell import InteractiveShell
from pympler import asizeof

from elastic.algorithm.baseline import MigrateAllBaseline, RecomputeAllBaseline
from elastic.algorithm.optimizer_exact import OptimizerExact
from elastic.algorithm.selector import OptimizerType
from elastic.core.common.profile_graph_size import profile_graph_size
from elastic.core.common.profile_migration_speed import profile_migration_speed
from elastic.core.graph.graph import DependencyGraph
from elastic.core.io.recover import resume
from elastic.core.mutation.fingerprint import compare_fingerprint, construct_fingerprint
from elastic.core.mutation.object_hash import UnserializableObj
from elastic.core.notebook.checkpoint import checkpoint
from elastic.core.notebook.find_input_vars import find_input_vars
from elastic.core.notebook.find_output_vars import find_created_deleted_vars
from elastic.core.notebook.restore_notebook import restore_notebook
from elastic.core.notebook.update_graph import update_graph


class ElasticNotebook:
    """
    Magics class for Elastic Notebook. Enable this in the notebook by running '%load_ext ElasticNotebook'.
    Enables efficient checkpointing of intermediate notebook state via balancing migration and recomputation
    costs.
    """

    def __init__(self, shell: InteractiveShell):
        self.shell = shell

        # Initialize the dependency graph for capturing notebook state.
        self.dependency_graph = DependencyGraph()

        # Migration properties.
        self.migration_speed_bps = 100000
        self.alpha = 1
        self.selector = OptimizerExact(migration_speed_bps=self.migration_speed_bps)

        # Dictionary of object fingerprints. For detecting modified references.
        self.fingerprint_dict = {}

        # Set of user-declared functions.
        self.udfs = set()

        # Flag if migration speed has been manually set. In this case, skip profiling of migration speed at checkpoint
        # time.
        self.manual_migration_speed = False

        # Location to log runtimes to. For experiments only.
        self.write_log_location = None

        # Strings for determining log filename. For experiments only.
        self.optimizer_name = ""
        self.notebook_name = ""

        # Total elapsed time spent inferring cell inputs and outputs.
        # For measuring overhead.
        self.total_recordevent_time = 0

        # Dict for recording overhead of profiling operations.
        self.profile_dict = {"idgraph": 0, "representation": 0}

        # デバッグフラグ
        self.debug = False

        # マイグレーションと再計算の変数リスト
        self._vss_to_migrate = []
        self._vss_to_recompute = []

    @property
    def vss_to_migrate(self):
        """マイグレーション対象の変数リストを取得"""
        return self._vss_to_migrate

    @property
    def vss_to_recompute(self):
        """再計算対象の変数リストを取得"""
        return self._vss_to_recompute

    def update_migration_lists(self, vss_to_migrate, vss_to_recompute):
        """マイグレーションと再計算の変数リストを更新"""
        self._vss_to_migrate = [vs.name for vs in vss_to_migrate]
        self._vss_to_recompute = [vs.name for vs in vss_to_recompute]

    def __str__(self):
        """文字列表現を定義"""
        return f"マイグレーション対象: {self.vss_to_migrate}\n再計算対象: {self.vss_to_recompute}"

    def set_debug(self, debug):
        self.debug = debug

    def record_event(self, cell):
        if self.debug:
            print("Recording event...")

        pre_execution = set(self.shell.user_ns.keys())

        # Create id trees for output variables
        for var in self.dependency_graph.variable_snapshots.keys():
            if var not in self.fingerprint_dict and var in self.shell.user_ns:
                self.fingerprint_dict[var] = construct_fingerprint(
                    self.shell.user_ns[var], self.profile_dict
                )

        # Find input variables (variables potentially accessed) of the cell.
        input_variables, function_defs = find_input_vars(
            cell,
            set(self.dependency_graph.variable_snapshots.keys()),
            self.shell,
            self.udfs,
        )
        # Union of ID graphs of input variables. For detecting modifications to unserializable variables.
        input_variables_id_graph_union = set()
        for var in input_variables:
            if var in self.fingerprint_dict:
                input_variables_id_graph_union = input_variables_id_graph_union.union(
                    self.fingerprint_dict[var][1]
                )

        # Run the cell.
        start_time = time.time()
        try:
            cell_output = get_ipython().run_cell(cell)
            cell_output.raise_error()
            # traceback_list = []
        except Exception:
            pass
            # _, _, tb = sys.exc_info()
            # traceback_list = traceback.extract_tb(tb).format()
        cell_runtime = time.time() - start_time
        post_execution = set(self.shell.user_ns.keys())
        infer_start = time.time()

        # Find created and deleted variables by computing difference between namespace pre and post execution.
        created_variables, deleted_variables = find_created_deleted_vars(
            pre_execution, post_execution
        )

        # Remove stored ID graphs for deleted variables.
        for var in deleted_variables:
            del self.fingerprint_dict[var]
            if var in self.udfs:
                self.udfs.remove(var)

        # Find modified variables by comparing ID graphs and object hashes.
        modified_variables = set()
        for k, v in self.fingerprint_dict.items():
            changed, overwritten = compare_fingerprint(
                self.fingerprint_dict[k],
                self.shell.user_ns[k],
                self.profile_dict,
                input_variables_id_graph_union,
            )
            if changed:
                modified_variables.add(k)

            # In the case of non-overwrite modification, the variable is additionally considered as accessed.
            if changed and not overwritten:
                input_variables.add(k)

            # A user defined function has been overwritten.
            elif overwritten and k in self.udfs:
                self.udfs.remove(k)

            # Select unserializable variables are assumed to be modified if accessed.
            if (
                not changed
                and not overwritten
                and isinstance(self.fingerprint_dict[k][2], UnserializableObj)
            ):
                if self.fingerprint_dict[k][1].intersection(
                    input_variables_id_graph_union
                ):
                    modified_variables.add(k)

        # Create ID graphs for output variables
        for var in created_variables:
            self.fingerprint_dict[var] = construct_fingerprint(
                self.shell.user_ns[var], self.profile_dict
            )

        # Record newly defined UDFs
        for udf in function_defs:
            if udf in self.shell.user_ns and isinstance(
                self.shell.user_ns[udf], types.FunctionType
            ):
                self.udfs.add(udf)

        # Update the dependency graph.
        update_graph(
            cell,
            cell_runtime,
            start_time,
            input_variables,
            created_variables.union(modified_variables),
            deleted_variables,
            self.dependency_graph,
        )

        # Update total recordevent time tally.
        infer_end = time.time()
        self.total_recordevent_time += infer_end - infer_start

    def set_migration_speed(self, migration_speed):
        try:
            if float(migration_speed) > 0:
                self.migration_speed_bps = float(migration_speed)
                self.manual_migration_speed = True
            else:
                pass
                # print("Migration speed is not positive.")
        except ValueError:
            print("Migration speed is not a number.")
        self.selector.migration_speed_bps = self.migration_speed_bps

    def set_optimizer(self, optimizer):
        self.optimizer_name = optimizer

        if optimizer == OptimizerType.EXACT.value:
            self.selector = OptimizerExact(self.migration_speed_bps)
            self.alpha = 1
        elif optimizer == OptimizerType.EXACT_C.value:
            self.selector = OptimizerExact(self.migration_speed_bps)
            self.alpha = 20
        elif optimizer == OptimizerType.EXACT_R.value:
            self.selector = OptimizerExact(self.migration_speed_bps)
            self.alpha = 0.05
        elif optimizer == OptimizerType.MIGRATE_ALL.value:
            self.selector = MigrateAllBaseline(self.migration_speed_bps)
        elif optimizer == OptimizerType.RECOMPUTE_ALL.value:
            self.selector = RecomputeAllBaseline(self.migration_speed_bps)

    def set_write_log_location(self, filename):
        self.write_log_location = filename

    def set_notebook_name(self, name):
        self.notebook_name = name

    def checkpoint(self, filename):
        """チェックポイントを作成"""
        if self.debug:
            print("Checkpointing...")

        # Write overhead metrics to file (for experiments).
        if self.write_log_location:
            with open(
                self.write_log_location
                + "/output_"
                + self.notebook_name
                + "_"
                + self.optimizer_name
                + ".txt",
                "a",
            ) as f:
                f.write(
                    "comparison overhead - "
                    + repr(
                        asizeof.asizeof(self.dependency_graph)
                        + asizeof.asizeof(self.fingerprint_dict)
                    )
                    + " bytes"
                    + "\n"
                )
                f.write(
                    "notebook overhead - "
                    + repr(asizeof.asizeof(self.shell.user_ns))
                    + " bytes"
                    + "\n"
                )
                f.write(
                    "Dependency graph storage overhead - "
                    + repr(profile_graph_size(self.dependency_graph))
                    + " bytes"
                    + "\n"
                )
                f.write(
                    "Cell monitoring overhead - "
                    + repr(self.total_recordevent_time)
                    + " seconds"
                    + "\n"
                )

        # Profile the migration speed to filename.
        if not self.manual_migration_speed:
            self.migration_speed_bps = profile_migration_speed(
                dirname(filename), alpha=self.alpha
            )
            self.selector.migration_speed_bps = self.migration_speed_bps

        # Checkpoint the notebook.
        return checkpoint(
            self.dependency_graph,
            self.shell,
            self.fingerprint_dict,
            self.selector,
            self.udfs,
            filename,
            self.profile_dict,
            self.write_log_location,
            self.notebook_name,
            self.optimizer_name,
            self,  # 自身のインスタンスを渡す
        )

    def load_checkpoint(self, filename):
        if self.debug:
            print("Loading checkpoint...")

        # start_time = time.time()
        (
            self.dependency_graph,
            variables,
            vss_to_migrate,
            vss_to_recompute,
            oes_to_recompute,
            self.udfs,
        ) = resume(
            filename, self.write_log_location, self.notebook_name, self.optimizer_name
        )

        # Recompute missing VSs and redeclare variables into the kernel.
        restore_notebook(
            self.dependency_graph,
            self.shell,
            variables,
            oes_to_recompute,
            self.write_log_location,
            self.notebook_name,
            self.optimizer_name,
        )
        # print("Checkpoint load time:", time.time() - start_time)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 University of Illinois
import time
from collections import defaultdict
from pathlib import Path

import dill

from elastic.core.io.filesystem_adapter import FilesystemAdapter
from elastic.core.io.migrate import FILENAME


def resume(
    filename: str, write_log_location=None, notebook_name=None, optimizer_name=None
):
    """
    Reads the file at `filename` and unpacks the graph representation of the notebook, migrated variables, and
    instructions for recomputation.

    Args:
        filename (str): Location of the checkpoint file.

        write_log_location (str): location to write file read runtime to. For experimentation only.
        notebook_name (str): notebook name. For experimentation only.
        optimizer_name (str): optimizer name. For experimentation only.
    """

    # Reads from the default location if a file path isn't specified.
    adapter = FilesystemAdapter()

    load_start = time.time()
    if filename:
        read_path = filename
    else:
        read_path = FILENAME

    variables = defaultdict(list)

    with open(Path(read_path), "rb") as output_file:
        metadata = dill.load(output_file)
        print(metadata.get_serialization_order())
        for vs_list in metadata.get_serialization_order():
            try:
                print("trying:", [vs.name for vs in vs_list])
                obj_list = dill.load(output_file)
                for i in range(len(vs_list)):
                    print("me: ", vs_list[i].output_ce.cell_num)
                    variables[vs_list[i].output_ce.cell_num].append(
                        (vs_list[i], obj_list[i])
                    )
            except Exception:
                print("oops")
                # unpickling failed. Rerun cells to retrieve variable(s).
                for vs in vs_list:
                    metadata.ces_to_recompute = metadata.ces_to_recompute.union(
                        metadata.recomputation_ces(vs.output_ce)
                    )

    if filename:
        metadata = adapter.read_all(Path(filename))
    else:
        metadata = adapter.read_all(Path(FILENAME))
    load_end = time.time()

    if write_log_location:
        with open(
            write_log_location
            + "/output_"
            + notebook_name
            + "_"
            + optimizer_name
            + ".txt",
            "a",
        ) as f:
            f.write("Reload stage took - " + repr(load_end - load_start) + " seconds\n")

    return (
        metadata.get_dependency_graph(),
        variables,
        metadata.get_vss_to_migrate(),
        metadata.get_vss_to_recompute(),
        metadata.get_ces_to_recompute(),
        metadata.get_udfs(),
    )



from elastic.core.graph.graph import DependencyGraph


def update_graph(
    cell: str,
    cell_runtime: float,
    start_time: float,
    input_variables: set,
    created_and_modified_variables: set,
    deleted_variables: set,
    graph: DependencyGraph,
):
    """
    Updates the graph according to the newly executed cell and its input and output variables.
    Args:
         cell (str): Raw cell cell.
         cell_runtime (float): Cell runtime.
         start_time (time): Time of start of cell execution. Note that this is different from when the cell was
             queued.
         input_variables (set): Set of input variables of the cell.
         created_and_modified_variables (set): set of created and modified variables.
         deleted_variables (set): set of deleted variables
         graph (DependencyGraph): Dependency graph representation of the notebook.
    """

    # Retrieve input variable snapshots
    input_vss = set(
        graph.variable_snapshots[variable][-1] for variable in input_variables
    )

    # Create output variable snapshots
    output_vss_create = set(
        graph.create_variable_snapshot(k, False) for k in created_and_modified_variables
    )
    output_vss_delete = set(
        graph.create_variable_snapshot(k, True) for k in deleted_variables
    )

    # Add the newly created OE to the graph.
    graph.add_cell_execution(
        cell,
        cell_runtime,
        start_time,
        input_vss,
        output_vss_create.union(output_vss_delete),
    )



import time

from ipykernel.zmqshell import ZMQInteractiveShell
from IPython import get_ipython
from IPython.utils.capture import capture_output

from elastic.core.graph.graph import DependencyGraph


def restore_notebook(
    graph: DependencyGraph,
    shell: ZMQInteractiveShell,
    variables: dict,
    ces_to_recompute: set,
    write_log_location=None,
    notebook_name=None,
    optimizer_name=None,
):
    """
    Restores the notebook. Declares variables back into the kernel and recomputes the CEs to restore non-migrated
    variables.
    Args:
        graph (DependencyGraph): dependency graph representation of the notebook.
        shell (ZMQInteractiveShell): interactive Jupyter shell storing the state of the current session.
        variables (Dict): Mapping from OEs to lists of variables defined in those OEs.
        oes_to_recompute (set): OEs to recompute to restore non-migrated variables.
        write_log_location (str): location to write component runtimes to. For experimentation only.
        notebook_name (str): notebook name. For experimentation only.
        optimizer_name (str): optimizer name. For experimentation only.
    """

    # Recompute OEs following the order they were executed in.
    recompute_start = time.time()
    for ce in graph.cell_executions:
        if ce in ces_to_recompute:
            # Rerun cell code; suppress stdout when rerunning.
            print("Rerunning cell", ce.cell_num + 1)
            cell_capture = capture_output(stdout=True, stderr=True, display=True)
            try:
                with cell_capture:
                    get_ipython().run_cell(ce.cell)
            except Exception as e:
                raise e

        # Define output variables in the CE.
        for pair in variables[ce.cell_num]:
            print("Declaring variable", pair[0].name)
            shell.user_ns[pair[0].name] = pair[1]

    recompute_end = time.time()

    if write_log_location:
        with open(
            write_log_location
            + "/output_"
            + notebook_name
            + "_"
            + optimizer_name
            + ".txt",
            "a",
        ) as f:
            f.write(
                "Recompute stage took - "
                + repr(recompute_end - recompute_start)
                + " seconds"
                + "\n"
            )



# /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 University of Illinois

import hashlib

# import polars as pl
import inspect
import mmap
import os
import pickle
import time
import types

import dill
import matplotlib.pyplot as plt
import networkx
import pandas as pd
import seaborn
from scipy import sparse


def is_picklable_fast(obj):
    if type(obj) in {types.GeneratorType, mmap.mmap, hashlib.sha256}:
        return False
    return True


def is_picklable(obj):
    """
    Checks whether an object is pickleable.
    """
    if is_exception(obj) or inspect.ismodule(obj):
        return True
    try:
        # This function can crash.
        return _is_picklable_dill(obj)
    except Exception:
        try:
            # Double check with function from pickle module
            return _is_picklable_raw(obj)
        except Exception:
            return False


def is_exception(obj):
    """
    List of objects which _is_picklable_dill returns false (or crashes) but are picklable.
    """
    if hasattr(obj, "__module__") and getattr(obj, "__module__", None).split(".")[
        0
    ] in {plt.__name__, seaborn.__name__, networkx.__name__, pd.__name__}:
        return True
    exceptions = [pd.core.frame.DataFrame, sparse.csr.csr_matrix]
    return type(obj) in exceptions


def _is_picklable_raw(obj):
    try:
        # dumps can be slow for large objects that can be pickled
        pickle.dumps(obj)
    except Exception:
        return False
    return True


def _is_picklable_dill(obj):
    # compared to _is_picklable_raw, this may be faster
    # however, dill's correctness is worrying because
    #   it currently considers Pandas DataFrame as not
    #   picklable, which is not true
    return dill.pickles(obj)
